"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const potentialInfinteLoopError_1 = __importDefault(require("./errors/potentialInfinteLoopError"));
const twoManyInstructionError_1 = __importDefault(require("./errors/twoManyInstructionError"));
const UnknownInstructionOrCellId_1 = __importDefault(require("./errors/UnknownInstructionOrCellId"));
const instructionsExecutor_1 = require("./instructions/instructionsExecutor");
const registers_1 = require("./registers");
class Sprint {
    maxInstructionCount;
    cells;
    inputReader;
    pc;
    insturctionsExecuted;
    isHalted = false;
    stepExecutionResult;
    constructor(maxInstructionCount, registers, inputReader, programCounter = 1, instructionExecuted = 0, stepExecutionResult = [], isHalted = false) {
        this.maxInstructionCount = maxInstructionCount;
        this.cells = registers;
        this.pc = programCounter;
        this.insturctionsExecuted = instructionExecuted;
        this.stepExecutionResult = stepExecutionResult;
        this.isHalted = isHalted;
        this.inputReader = inputReader;
    }
    static getInstance(maxInstructionCount, cellCount, code, inputReader) {
        return new Sprint(maxInstructionCount, Sprint.parse(updateLablesWithCellPositions(code), cellCount), inputReader);
    }
    nextInstructionLength() {
        if (this.isHalted) {
            return 0;
        }
        return (0, instructionsExecutor_1.getInstructionLength)(this.cells, this.pc);
    }
    executeNext() {
        if (this.insturctionsExecuted > this.maxInstructionCount) {
            throw new potentialInfinteLoopError_1.default(this.maxInstructionCount);
        }
        if (this.isHalted) {
            return;
        }
        const executionResult = (0, instructionsExecutor_1.executeInstruction)(this.cells.copy(), this.pc, this.inputReader);
        return new Sprint(this.maxInstructionCount, executionResult.cells.copy(), this.inputReader, executionResult.pc, this.insturctionsExecuted + 1, [...this.stepExecutionResult, { ...executionResult, previousCells: this.cells, previousPc: this.pc }], executionResult.isHalted);
    }
    execute() {
        while (!this.isHalted) {
            if (this.insturctionsExecuted > this.maxInstructionCount) {
                throw new potentialInfinteLoopError_1.default(this.maxInstructionCount);
            }
            const executionResult = (0, instructionsExecutor_1.executeInstruction)(this.cells.copy(), this.pc, this.inputReader);
            this.cells = executionResult.cells.copy();
            this.pc = executionResult.pc;
            this.insturctionsExecuted++;
            this.stepExecutionResult.push({ ...executionResult, previousCells: this.cells, previousPc: this.pc });
            this.isHalted = executionResult.isHalted;
        }
        return this;
    }
    executePrevious() {
        if (this.pc === 1) {
            return;
        }
        const lastExecutionResult = this.stepExecutionResult[this.stepExecutionResult.length - 1];
        return new Sprint(this.maxInstructionCount, lastExecutionResult.previousCells.copy(), this.inputReader, lastExecutionResult.previousPc, this.insturctionsExecuted - 1, this.stepExecutionResult.slice(0, this.stepExecutionResult.length - 1));
    }
    static parse(code, cellsCount) {
        const cells = new Array(cellsCount).fill(0)
            .map((_, i) => ({ id: i, value: undefined }));
        const tokens = code.split("\n")
            .flatMap((line) => line.trim().split(" "))
            .map(Sprint.toInt);
        if (tokens.length > cellsCount) {
            throw new twoManyInstructionError_1.default(tokens.length, cellsCount);
        }
        tokens.forEach((value, i) => {
            const cell = cells[i];
            if (cell)
                cell.value = value;
        });
        return new registers_1.Cells(cellsCount, cells);
    }
    static toInt(token) {
        const int = parseInt(token);
        if (isNaN(int)) {
            throw new UnknownInstructionOrCellId_1.default(token);
        }
        return int;
    }
}
exports.default = Sprint;
const ignoreComment = (line) => {
    if (line.includes(";")) {
        return line.substring(0, line.indexOf(";"));
    }
    return line;
};
const addLableIfNotPresent = (labels, label, position, failOnDuplicate) => {
    if (labels.get(label) && failOnDuplicate) {
        throw new Error(`Lable ${label} already declared for cell number ${labels.get(label)}`);
    }
    labels.set(label, position);
};
const getLabelsAndTokens = (rawCode, failOnDuplicate = true) => {
    const context = {
        labels: new Map(),
        cellsCount: 1,
        tokensWithOutLableDeclaration: []
    };
    rawCode.split("\n")
        .map(ignoreComment)
        .filter(line => line !== "")
        .reduce((acc, line) => {
        const tokens = line.replaceAll(":", ": ").split(/\s+/).filter(token => token !== "");
        tokens.forEach(token => {
            if (token.includes(":")) {
                const label = token.replace(":", "").trim();
                addLableIfNotPresent(acc.labels, label, acc.cellsCount, failOnDuplicate);
            }
            else {
                acc.cellsCount += 1;
            }
        });
        tokens.filter(token => !token.includes(":")).forEach(token => acc.tokensWithOutLableDeclaration.push(token));
        return acc;
    }, context);
    return context;
};
const updateLabelsWithCellPositions = (context) => {
    for (const entry of context.labels.entries()) {
        const updatedTokens = context.tokensWithOutLableDeclaration
            .map(token => token === entry[0] ? (entry[1] || "").toString() : token);
        context.tokensWithOutLableDeclaration = updatedTokens;
    }
    return context;
};
const validateNoUndeclaredVariableIsUsed = (context) => {
    context.tokensWithOutLableDeclaration.forEach((token, index) => {
        if (isNaN(parseInt(token))) {
            throw new Error(`Use of undeclared label ${token} at cell number ${index + 1}`);
        }
    });
    return context;
};
const updateLablesWithCellPositions = (rawCode) => {
    const context = getLabelsAndTokens(rawCode);
    updateLabelsWithCellPositions(context);
    validateNoUndeclaredVariableIsUsed(context);
    return context.tokensWithOutLableDeclaration.join(" ");
};
//# sourceMappingURL=sprint.js.map